package java.io;

import java.util.stream.StreamSupport;
import java.util.Spliterators;
import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.stream.Stream;

public class BufferedReader extends Reader
{
    private Reader in;
    private char[] cb;
    private int nChars;
    private int nextChar;
    private static final int INVALIDATED = -2;
    private static final int UNMARKED = -1;
    private int markedChar;
    private int readAheadLimit;
    private boolean skipLF;
    private boolean markedSkipLF;
    private static int defaultCharBufferSize;
    private static int defaultExpectedLineLength;
    
    public BufferedReader(Reader in, int n) {
        super(in);
        this.markedChar = -1;
        this.readAheadLimit = 0;
        this.skipLF = false;
        this.markedSkipLF = false;
        if (n <= 0) {
            throw new IllegalArgumentException("Buffer size <= 0");
        }
        this.in = in;
        this.cb = new char[n];
        int n2 = 0;
        this.nChars = n2;
        this.nextChar = n2;
        return;
    }
    
    public BufferedReader(Reader reader) {
        this(reader, BufferedReader.defaultCharBufferSize);
    }
    
    private void ensureOpen() throws IOException {
        if (this.in == null) {
            throw new IOException("Stream closed");
        }
        return;
    }
    
    private void fill() throws IOException {
        int nextChar;
        if (this.markedChar <= -1) {
            nextChar = 0;
        }
        else {
            int n = this.nextChar - this.markedChar;
            if (n >= this.readAheadLimit) {
                this.markedChar = -2;
                this.readAheadLimit = 0;
                nextChar = 0;
            }
            else {
                if (this.readAheadLimit <= this.cb.length) {
                    System.arraycopy(this.cb, this.markedChar, this.cb, 0, n);
                    this.markedChar = 0;
                    nextChar = n;
                }
                else {
                    char[] cb = new char[this.readAheadLimit];
                    System.arraycopy(this.cb, this.markedChar, cb, 0, n);
                    this.cb = cb;
                    this.markedChar = 0;
                    nextChar = n;
                }
                int n2 = n;
                this.nChars = n2;
                this.nextChar = n2;
            }
        }
        int i;
        do {
            i = this.in.read(this.cb, nextChar, this.cb.length - nextChar);
        } while (i == 0);
        if (i > 0) {
            this.nChars = nextChar + i;
            this.nextChar = nextChar;
        }
    }
    
    @Override
    public int read() throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            while (true) {
                if (this.nextChar >= this.nChars) {
                    this.fill();
                    if (this.nextChar >= this.nChars) {
                        return -1;
                    }
                }
                if (this.skipLF) {
                    this.skipLF = false;
                    if (this.cb[this.nextChar] == '\n') {
                        ++this.nextChar;
                    }
                    else {
                        break;
                    }
                }
                else {
                    break;
                }
            }
            return this.cb[this.nextChar++];
        }
    }
    
    private int read1(char[] array, int n, int a) throws IOException {
        if (this.nextChar >= this.nChars) {
            if (a >= this.cb.length && this.markedChar <= -1 && !this.skipLF) {
                return this.in.read(array, n, a);
            }
            this.fill();
        }
        if (this.nextChar >= this.nChars) {
            return -1;
        }
        if (this.skipLF) {
            this.skipLF = false;
            if (this.cb[this.nextChar] == '\n') {
                ++this.nextChar;
                if (this.nextChar >= this.nChars) {
                    this.fill();
                }
                if (this.nextChar >= this.nChars) {
                    return -1;
                }
            }
        }
        int min = Math.min(a, this.nChars - this.nextChar);
        System.arraycopy(this.cb, this.nextChar, array, n, min);
        this.nextChar += min;
        return min;
    }
    
    @Override
    public int read(char[] array, int n, int n2) throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            if (n < 0 || n > array.length || n2 < 0 || n + n2 > array.length || n + n2 < 0) {
                throw new IndexOutOfBoundsException();
            }
            if (n2 == 0) {
                return 0;
            }
            int read1 = this.read1(array, n, n2);
            if (read1 <= 0) {
                return read1;
            }
            while (read1 < n2 && this.in.ready()) {
                int read2 = this.read1(array, n + read1, n2 - read1);
                if (read2 <= 0) {
                    break;
                }
                read1 += read2;
            }
            return read1;
        }
    }
    
    String readLine(boolean b) throws IOException {
        Object o = null;
        synchronized (this.lock) {
            this.ensureOpen();
            int n = (b || this.skipLF) ? 1 : 0;
            while (true) {
                if (this.nextChar >= this.nChars) {
                    this.fill();
                }
                if (this.nextChar >= this.nChars) {
                    if (o != null && o.length() > 0) {
                        return o.toString();
                    }
                    return null;
                }
                else {
                    boolean n2 = false;
                    int n3 = '\0';
                    if (n != 0 && this.cb[this.nextChar] == '\n') {
                        ++this.nextChar;
                    }
                    this.skipLF = false;
                    n = 0;
                    int i = this.nextChar;
                    while (i < this.nChars) {
                        n3 = this.cb[i];
                        if (n3 == '\n' || n3 == '\r') {
                            n2 = true;
                            break;
                        }
                        ++i;
                    }
                    int nextChar = this.nextChar;
                    this.nextChar = i;
                    if (!(n2 == 0)) {
                        Object string;
                        if (o == null) {
                            string = new String(this.cb, nextChar, i - nextChar);
                        }
                        else {
                            o.append(this.cb, nextChar, i - nextChar);
                            string = o.toString();
                        }
                        ++this.nextChar;
                        if (n3 == '\r') {
                            this.skipLF = true;
                        }
                        return string;
                    }
                    else {
                        if (o == null) {
                            o = new StringBuilder(BufferedReader.defaultExpectedLineLength);
                        }
                        o.append(this.cb, nextChar, i - nextChar);
                    }
                }
            }
        }
    }
    
    public String readLine() throws IOException {
        return this.readLine(false);
    }
    
    @Override
    public long skip(long n) throws IOException {
        if (n < 0L) {
            throw new IllegalArgumentException("skip value is negative");
        }
        synchronized (this.lock) {
            this.ensureOpen();
            long n2 = n;
            while (n2 > 0L) {
                if (this.nextChar >= this.nChars) {
                    this.fill();
                }
                if (this.nextChar >= this.nChars) {
                    break;
                }
                if (this.skipLF) {
                    this.skipLF = false;
                    if (this.cb[this.nextChar] == '\n') {
                        ++this.nextChar;
                    }
                }
                long n3 = (long)(this.nChars - this.nextChar);
                if (n2 <= n3) {
                    this.nextChar = (int)((long)this.nextChar + n2);
                    n2 = 0L;
                    break;
                }
                n2 -= n3;
                this.nextChar = this.nChars;
            }
            return n - n2;
        }
    }
    
    @Override
    public boolean ready() throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            if (this.skipLF) {
                if (this.nextChar >= this.nChars && this.in.ready()) {
                    this.fill();
                }
                if (this.nextChar < this.nChars) {
                    if (this.cb[this.nextChar] == '\n') {
                        ++this.nextChar;
                    }
                    this.skipLF = false;
                }
            }
            return this.nextChar < this.nChars || this.in.ready();
        }
    }
    
    @Override
    public boolean markSupported() {
        return true;
    }
    
    @Override
    public void mark(int readAheadLimit) throws IOException {
        if (readAheadLimit < 0) {
            throw new IllegalArgumentException("Read-ahead limit < 0");
        }
        synchronized (this.lock) {
            this.ensureOpen();
            this.readAheadLimit = readAheadLimit;
            this.markedChar = this.nextChar;
            this.markedSkipLF = this.skipLF;
        }
        return;
    }
    
    @Override
    public void reset() throws IOException {
        synchronized (this.lock) {
            this.ensureOpen();
            if (this.markedChar < 0) {
                throw new IOException((this.markedChar == -2) ? "Mark invalid" : "Stream not marked");
            }
            this.nextChar = this.markedChar;
            this.skipLF = this.markedSkipLF;
        }
    }
    
    @Override
    public void close() throws IOException {
        synchronized (this.lock) {
            if (this.in == null) {
                return;
            }
            try {
                this.in.close();
            }
            finally {
                this.in = null;
                this.cb = null;
            }
        }
    }
    
    public Stream<String> lines() {
        return StreamSupport.stream(Spliterators.spliteratorUnknownSize(new Iterator<String>() {
            String nextLine = null;
            
            @Override
            public boolean hasNext() {
                if (this.nextLine != null) {
                    return true;
                }
                try {
                    this.nextLine = BufferedReader.this.readLine();
                    return this.nextLine != null;
                }
                catch (IOException ex) {
                    throw new UncheckedIOException(ex);
                }
            }
            
            @Override
            public String next() {
                if (this.nextLine != null || this.hasNext()) {
                    String nextLine = this.nextLine;
                    this.nextLine = null;
                    return nextLine;
                }
                throw new NoSuchElementException();
            }
        }, 272), false);
    }
    
    static {
        BufferedReader.defaultCharBufferSize = 8192;
        BufferedReader.defaultExpectedLineLength = 80;
    }
}
